import logging
from unittest import TestCase
import mxnet as mx
import numpy as np
from mxnet_octave_conv import octave_conv, octave_dual, octave_residual


class TestMXNetOctave(TestCase):

    def _test_fit(self, symbol):
        logging.getLogger().setLevel(logging.DEBUG)
        data_size = 4096
        batch_size = 32
        x = np.random.standard_normal((data_size, 3, 32, 32))
        y = np.random.randint(0, 2, data_size)
        data = mx.io.NDArrayIter(x, y, batch_size=batch_size)
        model = mx.model.FeedForward(
            symbol=symbol,
            num_epoch=5,
            optimizer='adam',
            numpy_batch_size=batch_size,
        )
        model.fit(
            X=data,
            batch_end_callback=mx.callback.Speedometer(batch_size, 200),
        )
        predict = model.predict(x).argmax(axis=-1)
        self.assertLess(np.sum(np.abs(y - predict)), data_size // 2)

    def test_fit_basic(self):
        x = mx.symbol.Variable(name='data')
        conv = octave_conv(x, num_filter=7, kernel=(3, 3))
        pool = octave_dual(conv, lambda data: mx.symbol.Pooling(data, kernel=(2, 2), stride=(2, 2), pool_type='max'))
        conv = octave_conv(pool, num_filter=5, kernel=3, stride=1, dilate=(2, 3), name='Mid')
        pool = octave_dual(conv, lambda data: mx.symbol.Pooling(data, kernel=(2, 2), stride=(2, 2), pool_type='max'))
        conv = octave_conv(pool, num_filter=3, kernel=3, stride=(1, 1), dilate=1, ratio_out=0.0)
        pool = octave_dual(conv, lambda data: mx.symbol.Pooling(data, kernel=(2, 2), stride=(2, 2), pool_type='max'))
        flatten = mx.symbol.Flatten(pool)
        dense = mx.symbol.FullyConnected(flatten, num_hidden=2)
        model = mx.symbol.SoftmaxOutput(dense, name='softmax')
        print(mx.visualization.print_summary(model, shape={'data': (2, 3, 32, 32)}))
        self._test_fit(model)

    def test_fit_low(self):
        x = mx.symbol.Variable(name='data')
        conv = octave_conv(x, num_filter=7, kernel=(3, 3))
        pool = octave_dual(conv, lambda data: mx.symbol.Pooling(data, kernel=(2, 2), stride=(2, 2), pool_type='max'))
        conv = octave_conv(pool, num_filter=5, kernel=3, stride=1, dilate=(2, 3), name='Mid')
        pool = octave_dual(conv, lambda data: mx.symbol.Pooling(data, kernel=(2, 2), stride=(2, 2), pool_type='max'))
        conv = octave_conv(pool, num_filter=3, kernel=3, stride=(1, 1), dilate=1, ratio_out=1.0)
        pool = octave_dual(conv, lambda data: mx.symbol.Pooling(data, kernel=(2, 2), stride=(2, 2), pool_type='max'))
        flatten = mx.symbol.Flatten(pool)
        dense = mx.symbol.FullyConnected(flatten, num_hidden=2)
        model = mx.symbol.SoftmaxOutput(dense, name='softmax')
        print(mx.visualization.print_summary(model, shape={'data': (2, 3, 32, 32)}))
        self._test_fit(model)

    def test_fit_residual(self):
        x = mx.symbol.Variable(name='data')
        conv = octave_conv(x, num_filter=8, kernel=(3, 3))
        conv_dual = octave_dual(conv, lambda data: mx.symbol.Convolution(data, num_filter=4, kernel=(3, 3), pad=(1, 1)))
        residual = octave_residual(conv, conv_dual)
        pool = octave_dual(
            residual,
            lambda data: mx.symbol.Pooling(data, kernel=(2, 2), stride=(2, 2), pool_type='max'),
        )
        conv = octave_conv(pool, num_filter=5, kernel=3, stride=1, dilate=(2, 3), name='Mid')
        pool = octave_dual(conv, lambda data: mx.symbol.Pooling(data, kernel=(2, 2), stride=(2, 2), pool_type='max'))
        conv = octave_conv(pool, num_filter=3, kernel=3, stride=(1, 1), dilate=1, ratio_out=1.0)
        pool = octave_dual(conv, lambda data: mx.symbol.Pooling(data, kernel=(2, 2), stride=(2, 2), pool_type='max'))
        flatten = mx.symbol.Flatten(pool)
        dense = mx.symbol.FullyConnected(flatten, num_hidden=2)
        model = mx.symbol.SoftmaxOutput(dense, name='softmax')
        print(mx.visualization.print_summary(model, shape={'data': (2, 3, 32, 32)}))
        self._test_fit(model)



import mxnet as mx

__all__ = ['octave_conv', 'octave_dual', 'octave_residual']


def _make_tuple(x, rank):
    if isinstance(x, int):
        return (x,) * rank
    return tuple(x)


def _make_name(name, suffix=''):
    if name is None:
        return None
    return name + '-' + suffix


def octave_conv(data,
                num_filter,
                kernel,
                octave=2,
                ratio_out=0.5,
                stride=1,
                dilate=1,
                num_group=1,
                workspace=1024,
                no_bias=0,
                cudnn_tune=None,
                cudnn_off=0,
                layout=None,
                name=None,
                attr=None,
                **kwargs):
    """Create octave convolution structure.

    Parameters
    ----------
    octave : int (positive), optional, default=2
        The division of the spatial dimensions by a power of 2.
    ratio_out: float (non-negative), optional, default=0.5
        The ratio of filters for lower spatial resolution.

    Returns
    -------
    Symbol(tuple)
        The result symbols. Only one symbol will be returned if ratio_out is either 0.0 or 1.0.

    References
    ----------
    - [Drop an Octave: Reducing Spatial Redundancy in Convolutional Neural Networks with Octave Convolution]
      (https://arxiv.org/pdf/1904.05049.pdf)
    """
    if isinstance(data, (list, tuple)):
        data_high, data_low = data
    else:
        data_high, data_low = data, None

    num_filter_low = int(num_filter * ratio_out)
    num_filter_high = num_filter - num_filter_low

    rank = 1
    if isinstance(kernel, (list, tuple)):
        rank = max(rank, len(kernel))
    if isinstance(stride, (list, tuple)):
        rank = max(rank, len(stride))
    if isinstance(dilate, (list, tuple)):
        rank = max(rank, len(dilate))
    kernel = _make_tuple(kernel, rank)
    stride = _make_tuple(stride, rank)
    dilate = _make_tuple(dilate, rank)
    octave = _make_tuple(octave, rank)
    pad = tuple([k // 2 * d for k, d in zip(kernel, dilate)])

    def _init_conv(_data, _num_filter, _suffix):
        return mx.symbol.Convolution(
            data=_data,
            kernel=kernel,
            stride=stride,
            dilate=dilate,
            pad=pad,
            num_filter=_num_filter,
            num_group=num_group,
            workspace=workspace,
            no_bias=no_bias,
            cudnn_tune=cudnn_tune,
            cudnn_off=cudnn_off,
            layout=layout,
            name=_make_name(name, _suffix),
            attr=attr,
            **kwargs
        )

    outputs_high = None
    if num_filter_high > 0:
        outputs_high = _init_conv(data_high, num_filter_high, 'hh')
        if data_low is not None:
            outputs_high = mx.symbol.elemwise_add(
                outputs_high,
                mx.symbol.UpSampling(
                    _init_conv(data_low, num_filter_high, 'lh'),
                    scale=octave[0],
                    sample_type='nearest',
                    name=_make_name(name, 'up'),
                ),
                name=_make_name(name, 'add-h'),
            )

    outputs_low = None
    if num_filter_low > 0:
        outputs_low = _init_conv(
            mx.symbol.Pooling(
                data_high,
                kernel=octave,
                pool_type='avg',
                stride=octave,
                cudnn_off=cudnn_off,
                name=_make_name(name, 'pooling'),
            ),
            num_filter_low, 'hl',
        )
        if data_low is not None:
            outputs_low = mx.symbol.elemwise_add(
                _init_conv(data_low, num_filter_low, 'll'),
                outputs_low,
                name=_make_name(name, 'add-l'),
            )

    if outputs_high is None:
        return outputs_low
    if outputs_low is None:
        return outputs_high
    return outputs_high, outputs_low


def octave_dual(data, builder):
    if not isinstance(data, (list, tuple)):
        data = [data]
    outputs = [builder(datum) for datum in data]
    if len(outputs) == 1:
        outputs = outputs[0]
    return outputs


def octave_residual(x, y):
    return octave_dual(list(zip(x, y)), lambda z: z[0] + z[1])
